import numpy as np
import itertools as it
import pandas as pd
import seaborn as sns
from .stratum import Stratum
from typing import Union, Hashable, Iterable, Optional, Any


class Normalizer:
    def __init__(
        self,
        vmin: float,
        vmax: float
    ):
        self.vmin = vmin
        self.vmax = vmax

    def __call__(
        self,
        val: float,
        scale: float = None
    ) -> float:
        norm = (val - self.vmin) / (self.vmax - self.vmin)
        return norm * scale if scale else norm


def get_color_dict(
    data: pd.DataFrame,
    x: str,
    stratum: str,
    hue: Optional[str],
    sns_palette: str = 'husl'
) -> Union[str, dict[Hashable, dict[Hashable, tuple[float, float, float, float]]]]:
    """
    returns a dict of colors as expected by alluvial

    :param data:            data to plot in wide format
    :param x:               column in data by which to group along x axis
    :param stratum:         column in data by which to group groups along x axis
    :param hue:             column in data indicating the grouping of strata within groups or None
    :param sns_palette:     string indicating the palette to use for colors (see also seaborn)

    :return:                dictionary of colors of the form {group_name: {stratum_name: color}}
    """

    if hue:
        group_sizes = [group.grouping.nunique() * 2 for _, group in groupby(data, x)]

    else:
        group_sizes = [group[stratum].nunique() for _, group in groupby(data, x)]

    palette = sns.color_palette(
        sns_palette,
        max(group_sizes)
    )

    color_dict = {}
    for group_name, strata in groupby(data, x):
        color_dict[group_name] = {}
        for i, (stratum_group_name, stratum_group) in enumerate(groupby(strata, 'grouping')):
            if hue:
                for j, (stratum_name,  _) in enumerate(groupby(stratum_group, stratum)):
                    color_dict[group_name][stratum_name] = palette[2*i + j]

            else:
                color_dict[group_name][stratum_group_name] = palette[i]

    return color_dict


def pairwise(iterable: Iterable) -> list[tuple[Any, Any]]:
    """
    returns successive overlapping pairs taken from the input iterable.
    taken from itertools since this is only available for python>=3.10

    :param iterable:    any iterable container
    :return:            successive overlapping pairs
    """

    a, b = it.tee(iterable)
    next(b, None)
    return list(zip(a, b))


def generate_probabilities(n: int) -> list[float]:
    """
    generates an array of n probabilities with the constraint of sum(probabilities) = 1

    :param n:   number of probabilities in the array
    :return:    list of floats
    """

    all_nonzero = False
    while not all_nonzero:
        probabilities = [np.random.randint(0, 51)]
        high = 101 - probabilities[0]
        for i in range(n - 2):
            p = np.random.randint(0, high)
            probabilities.append(p)
            high -= p

        probabilities.append(
            100 - sum(probabilities)
        )

        all_nonzero = all(probabilities)

    return [p / 100 for p in probabilities]


def to_dataframe(
    x: Iterable,
    alluvium: Iterable,
    stratum: Iterable,
    hue: Optional[Iterable] = None
) -> pd.DataFrame:
    """
    converts three iterables of the same length to a pandas.DataFrame
    :param x:
    :param alluvium:
    :param stratum:
    :return:
    """
    keys = ['x', 'alluvium', 'stratum']
    iterables = [x, alluvium, stratum]

    if not isinstance(hue, type(None)):
        keys += ['hue']
        iterables += [hue]

    df = pd.DataFrame(
        {k: v for k, v in zip(keys, iterables)}
    )

    return df


# groupby lambda expression for shorter code
groupby = lambda df, key: df.groupby(key, sort = True)


def reset_strata(strata: list[Stratum]) -> None:
    """
    resets the flow position tracking of all given Stratum objects

    :param strata:  list of Stratum object to reset

    :return:        None
    """
    for stratum in strata:
        stratum.reset_lode_position()


def generate_test_data(
        group_sizes: Union[tuple[int], list[int]] = (3, 5, 4, 6)
) -> pd.DataFrame:
    """
    generates test data for the alluvial plotter

    :param group_sizes:         list or tuple of integers specifying the number of strata per group to generate test data for

    :return: pandas.DataFrame
    """

    nnodes = 1000
    ngroups = len(group_sizes)
    nodenames = list(range(nnodes)) * ngroups
    timepoint = [f't{i}' for i in range(ngroups) for j in range(nnodes)]
    probabilities = [generate_probabilities(i) for i in group_sizes]
    modules, hue = [], []
    for p in probabilities:
        modules.append(
            np.random.choice(range(1, len(p) + 1), size = nnodes, p = p, replace = True)
        )
        phue = np.random.rand()
        hue.append(
            np.random.choice(['s', 'ns'], size = nnodes, p = [phue, 1 - phue])
        )

    modules = np.concatenate(modules)
    hue = np.concatenate(hue)

    df = pd.DataFrame(
        [[a, b, c, d] for a, b, c, d in zip(nodenames, timepoint, modules, hue)],
        columns = [
            'nodename',
            'timepoint',
            'module',
            'signif'
        ]
    )
    return df



import numpy as np

def poly_fit_with_straights(y1, y2, x1, x2, resolution, straight_fraction):
    '''
    fits a 4th grade polynomial between x1, y1 and x2, y2
    with leading and trailing straight lines

    :param y1:                  y coordinate of the first point
    :param y2:                  y coordinate of the second point
    :param x1:                  x coordinate of the first point
    :param x2:                  x coordinate of the second point
    :param resolution:          resolution of the interpolated polynomial
    :param straight_fraction:   fraction of the space between x1 and x2 that should be straight lines

    :return:                    numpy.ndarray, np.ndarray containing x and y coordinates of the fitted function
    '''
    # adapted from https://github.com/vinsburg/alluvial_diagram/blob/master/alluvial.py
    straight_length = (x2 - x1) * straight_fraction
    straight1_x = np.linspace(
        x1,
        x1 + straight_length,
        resolution
    )
    straight1_y = np.repeat(y1, resolution)
    straight2_x = np.linspace(
        x2 - straight_length,
        x2,
        resolution
    )
    straight2_y = np.repeat(y2, resolution)

    y = np.array([0, 0.15, 0.5, 0.85, 1])
    y = y * (y2 - y1) + y1
    x = np.linspace(
        x1 + straight_length,
        x2 - straight_length,
        len(y)
    )
    z = np.polyfit(x, y, 4)
    f = np.poly1d(z)

    xs = np.linspace(x[0], x[-1], resolution)
    ys = f(xs)

    xs = np.concatenate([straight1_x, xs, straight2_x])
    ys = np.concatenate([straight1_y, ys, straight2_y])

    return xs, ys

def sigmoid_fit(y1, y2, x1, x2, resolution):
    '''
    fits a sigmoidal curve between x1, y1 and x2, y2
    with leading and trailing straight lines

    :param y1:                  y coordinate of the first point
    :param y2:                  y coordinate of the second point
    :param x1:                  x coordinate of the first point
    :param x2:                  x coordinate of the second point
    :param resolution:          resolution of the interpolated sigmoidal

    :return:                    numpy.ndarray, np.ndarray containing x and y coordinates of the fitted function
    '''
    sigmoid = lambda x: 1/(1 + np.exp(-x))

    xs = np.linspace(-10, 10, resolution)
    ys = sigmoid(xs)

    xs = np.linspace(x1, x2, resolution)
    ys = ys * (y2 - y1) + y1

    return xs, ys


from matplotlib.patches import Polygon
from typing import Any, Union, Optional


class Stratum:
    def __init__(
        self,
        relative_height: float,
        x: float = 0,
        y: float = 0,
        label: Optional[Any] = None
):
        self.x = x
        self.y = y
        self.relative_height = relative_height
        self.lode_position = 0
        self.height = 0
        self.width = 0
        self.color = None
        self.label = label

    def __repr__(self) -> str:
        return f'Stratum(h = {self.height:.02f}, rh = {self.relative_height:.02f}, y = {self.y:.02f})'

    def set_width(self, width: float) -> None:
        self.width = width

    def set_color(self, color: Union[str, tuple[float, float, float, float]]) -> None:
        self.color = color

    def reset_lode_position(self) -> None:
        self.lode_position = self.height

    def get_left_bound(self, gap: float) -> float:
        return self.x - self.width / 2 - gap

    def get_right_bound(self, gap: float) -> float:
        return self.x + self.width / 2 + gap

    def get_flow_ycoords(self, relative_flow_width: float) -> tuple[float, float]:
        if not self.lode_position:
            self.lode_position = self.height

        flow_width = self.height * relative_flow_width
        top = self.lode_position + self.y
        bottom = top - flow_width
        self.lode_position -= flow_width

        return top, bottom

    def set_xy(self, x: float, y: float) -> None:
        self.x = x
        self.y = y

    def set_height(self, scale, norm = None):
        height = self.relative_height * scale
        self.height = norm(height, scale) if norm else height

    def get_patch(
        self,
        color: Union[str, tuple[float, float, float, float]],
        alpha: float
    ) -> Polygon:
        if not self.color:
            self.color = color

        top_left = [self.x - self.width / 2, self.y + self.height]
        top_right = [self.x + self.width / 2, self.y + self.height]
        bottom_left = [self.x - self.width / 2, self.y]
        bottom_right = [self.x + self.width / 2, self.y]
        patch = Polygon(
            [
                top_left, top_right,
                bottom_right, bottom_left
            ],
            facecolor = color,
            edgecolor = 'white',
            alpha = alpha
        )
        return patch

    def get_label(self) -> tuple[float, float, str]:
        return self.x, self.y + self.height / 2, str(self.label)



from .stratum import Stratum
from .utils import pairwise, groupby
import numpy as np
import pandas as pd
from typing import Any


def get_lodes(
    data: pd.DataFrame,
    x: str,
    alluvium: str,
    stratum: str
) -> list[list[np.ndarray]]:
    """
    computes lode widths for each flow between strata of successive groups in x
    returns a nested list of numpy arrays of the form l = list(list(array())), where
    l[i][j] gives the flows computed between groups i and i+1 for stratum j as a 2D array
    of shape n x 2, where n gives the number of flows from stratum j in group i to strata in
    group i+1 (e.g. if stratum in i has flows to three strata in i+1 n = 3). The two columns
    of this array denote the width of the lode as a fraction of stratum j in i (first column)
    or of stratum k in i+1 (second column)

    :param data:        pandas.DataFrame
    :param x:           categorical column on which to group data
    :param alluvium:    column on which to compute flows
    :param stratum:     categorical column on which to group x groups
    :param hue:         grouping variable on which to split each stratum

    :return:            nested list of numpy.ndarrays
    """

    lodes = []
    for (_, g1), (_, g2) in pairwise(groupby(data, x)):
        flows = []
        for _, g1_strat in groupby(g1, stratum):
            strat_flows = []
            for _, g2_strat in groupby(g2, stratum):
                flow = g2_strat[g2_strat[alluvium].isin(g1_strat[alluvium])]
                strat_flows.append(
                    [len(flow) / len(strat) for strat in [g1_strat, g2_strat]]
                )

            flows.append(
                np.array(strat_flows)
            )
        lodes.append(flows)

    return lodes


def aggregate_data(
    data: pd.DataFrame,
    x: str,
    alluvium: str,
    stratum: str
) -> tuple[list[list[Stratum]], list[list[np.ndarray]], list[Any], list[list[Any]]]:
    """
    computes the strata and lodes for each categorical column in x

    :param data:        pandas.DataFrame containing data in long format (see generate_test_data)
    :param x:           categorical column on which to split the data along the x axis
    :param alluvium:    column from which to compute lode sizes for each stratum
    :param stratum:     column specifying the strata for each column in x

    :return:            list of lists Stratum objects, lode sizes as list of lists of arrays with
                        l[i][j] = lodesizes relative to strata in i and i + 1
                        see also function `get_lodes` for more context
    """

    strata_by_group, group_labels, strata_groupings = [], [], []
    for group_label, group in groupby(data, x):
        group_labels.append(group_label)
        strata, strat_grouping = [], []
        for label, strat in groupby(group, stratum):
            strata.append(
                Stratum(len(strat) / len(group), 0, 0, label)
            )
            strat_grouping.append(
                strat.grouping.unique()
            )
        strata_by_group.append(strata)
        strata_groupings.append(strat_grouping)

    lodes = get_lodes(
        data,
        x,
        alluvium,
        stratum
    )

    return strata_by_group, lodes, group_labels, strata_groupings



from .utils import *
from .aggregate import aggregate_data
from .fit import *
from matplotlib.patches import Polygon
from typing import Hashable, Optional
import matplotlib.pyplot as plt
import numpy as np
import itertools as it


def get_flow_path(
    y1: float,
    y2: float,
    x1: float,
    x2: float,
    resolution: int = 50,
    straight_fraction: float = 0.2,
    fit: str = 'poly'
) -> tuple[np.ndarray, np.ndarray]:
    """
    compute the flow path between x1, y1 and x2, y2 using either
    a polynomial with leading and trailing straight lines or a sigmoidal

    :param y1:                  y coordinate of the first point
    :param y2:                  y coordinate of the second point
    :param x1:                  x coordinate of the first point
    :param x2:                  x coordinate of the second point
    :param resolution:          resolution of the interpolated function
    :param straight_fraction:   fraction of the space between x1 and x2 that should be straight lines for fit = 'poly'
    :param fit:                 string specifying the function to use for computing the flow path

    :return:                    numpy.ndarray, np.ndarray containing x and y coordinates of the fitted function
    """
    if fit == 'sigmoid':
        xs, ys = sigmoid_fit(
            y1, y2,
            x1, x2,
            resolution
        )

    elif fit == 'poly':
        xs, ys = poly_fit_with_straights(
            y1, y2,
            x1, x2,
            resolution,
            straight_fraction
        )

    return xs, ys


def make_flow_polygon(
    g1_strat: Stratum,
    g2_strat: Stratum,
    g1_rh: float,
    g2_rh: float,
    color: Union[str, tuple[float, float, float, float]],
    alpha: float = 1
) -> Polygon:
    """
    generates a matplotlib.patches.Polygon object for a given alluvial flow

    :param g1_strat:    origin Stratum object
    :param g2_strat:    destination Stratum object
    :param g1_rh:       flow proportion of origin Stratum
    :param g2_rh:       flow proportion of destination Stratum
    :param color:       color of the generated Polygon
    :param alpha:       opacity of the generated Polygon

    :return:            matplotlib.patches.Polygon for flow between origin and destination Stratum
    """
    y1_bottom, y1_top = g1_strat.get_flow_ycoords(g1_rh)
    y2_bottom, y2_top = g2_strat.get_flow_ycoords(g2_rh)
    x1 = g1_strat.get_right_bound(0.5)
    x2 = g2_strat.get_left_bound(0.5)
    x_bottom, y_bottom = get_flow_path(y1_bottom, y2_bottom, x1, x2)
    x_top, y_top = get_flow_path(y1_top, y2_top, x1, x2)
    x = np.concatenate([x_top, x_bottom[::-1]])
    y = np.concatenate([y_top, y_bottom[::-1]])
    flow_polygon = Polygon(
        np.array([x, y]).T,
        facecolor = color,
        edgecolor = 'white',
        alpha = alpha
    )

    return flow_polygon


def plot_group_strata(
    group_strata: list[Stratum],
    x: float,
    grouping: list[Any],
    colors: dict[Hashable, Union[str, tuple[float, float, float, float]]],
    ax: plt.Axes,
    gapsize: float = 1,
    height: float = 100,
    width: float = 0.5,
    alpha: float = 1,
    show_labels: bool = False
) -> None:
    """
    plots stratas for a given group

    :param group_strata:    list of Stratum objects
    :param x:               position of group column on x axis
    :param grouping:        list of integers indicating strata that should not have a gap between them
    :param colors:          dictionary of the form {stratum_name: color}
    :param ax:              matplotlib.Axes object to add the strata patches to
    :param gapsize:         size of gap between each pair of rectangles
    :param height:          height of the group column
    :param width:           width of the plotted rectangles
    :param alpha:           opacity of plotted rectangles
    :param show_labels:     if True, plots Stratum labels

    :return:                None
    """
    y = 0
    ngaps = len(np.unique(grouping))
    norm = Normalizer(y, height + gapsize * (ngaps - 1))
    # group_strata[::-1] is necessary to comply to top to bottom order of strata
    for i, strat_group in it.groupby(
            zip(group_strata[::-1], grouping[::-1]),
            key = lambda x: x[1]
    ):
        for stratum, _ in strat_group:
            c = colors[stratum.label]
            stratum.set_height(height, norm)
            stratum.set_width(width)
            stratum.set_xy(x, y)
            ax.add_patch(
                stratum.get_patch(c, alpha)
            )
            if show_labels:
                ax.text(
                    *stratum.get_label(),
                    rotation = 90,
                    ha = 'center',
                    va = 'center'
                )
            y += stratum.height

        y += gapsize


def plot_strata(
    strata: list[list[Stratum]],
    group_labels: list[str],
    groupings: list[list[Any]],
    strat_colors: dict[Hashable, dict[Hashable, Union[str, tuple[float, float, float, float]]]],
    stratum_width: float,
    stratum_gap: float,
    ax: plt.Axes,
    plot_height: float,
    plot_width: float,
    show_labels: bool
) -> None:
    """
    plots strata for each group

    :param strata:          list of lists of Stratum objects
    :param group_labels:    list of labels for each strata group
    :param groupings:       list of list of group labels indicating a grouping of the Stratum objects for each group
    :param strat_colors:    dictionary of the form {group_name: {stratum_name: color}}
    :param stratum_width:   width of the Stratum rectangles
    :param stratum_gap      size of the gap between strata
    :param ax:              matplotlib.Axes object to add the strata patches to
    :param plot_height:     height of the groups
    :param plot_width:      width of the plot
    :param show_labels:     if True, plots Stratum labels

    :return:                None
    """

    x = stratum_width / 2
    label_positions = []
    for group_label, group_strata, grouping in zip(group_labels, strata, groupings):
        colors = strat_colors[group_label]
        label_positions.append(x)
        plot_group_strata(
            group_strata,
            x,
            grouping,
            colors,
            ax,
            gapsize = stratum_gap,
            height = plot_height,
            width = stratum_width,
            show_labels = show_labels
        )
        x += plot_width / len(strata)

    if show_labels:
        ax.set_xticks(label_positions)
        ax.set_xticklabels(group_labels)

    ax.set_xlim(0, x - plot_width / len(strata) + stratum_width / 2)
    ax.set_ylim(0, plot_height)


def plot_flows(
    strata: list[list[Stratum]],
    lodes: list[list[np.ndarray]],
    ax: plt.Axes
) -> None:
    """
    plot flows between strata

    :param strata:  list of lists of Stratum objects
    :param lodes:   list of lists of numpy.ndarrays holding the Stratum flow proportions (see also get_lodes)
    :param ax:      matplotlib.Axes object to add the flows to

    :return:        None
    """
    for (i, g1_strats), (_, g2_strats) in pairwise(enumerate(strata)):
        for j, g1_strat in enumerate(g1_strats):
            relative_widths = lodes[i][j]
            for k, g2_strat in enumerate(g2_strats):
                flow_polygon = make_flow_polygon(
                    g1_strat,
                    g2_strat,
                    relative_widths[k][0],
                    relative_widths[k][1],
                    g1_strat.color
                )
                ax.add_patch(flow_polygon)

        # reset lode_position for next round
        reset_strata(g2_strats)


def alluvial(
    x: Union[str, Iterable],
    stratum: Union[str, Iterable],
    alluvium: Union[str, Iterable],
    palette: str = 'husl',
    hue: Optional[Union[str, Iterable]] = None,
    data: Optional[pd.DataFrame] = None,
    ax: Optional[plt.Axes] = None,
    stratum_width: float = 2,
    stratum_gap: float = 1,
    plot_height: float = 100,
    plot_width: float = 150,
    show_labels: bool = False
) -> Union[plt.Axes, tuple[plt.Figure, plt.Axes]]:
    """
    generate alluvial plot. x, stratum and alluvium are either strings if data is given or iterables of same length
    containing the data to plot in long format (e.g. [0, 0, 1, 1, 2, 2], [0, 1, 0, 1, 0, 1], [0, 1, 0, 1, 0, 1]).
    In the latter case hue also has to be an iterable of the same length

    :param x:               string denoting the column to use for grouping of data if data is given else iterable
    :param stratum:         string denoting the column to use for computing stata heights if data is given else iterable
    :param alluvium:        string denoting the column to use for computing flows between strata if data is given else iterable
    :param palette:         string denoting a given seaborn palette or dictionary of the form {'group_name': {'statum_name': 'color'}}
    :param hue:             string denoting the column to use for computing Stratum splits if data else iterable
    :param data:            pandas.DataFrame containing data in long format
    :param ax:              matplotlib.Axes object to generate the plot in
    :param stratum_width:   width of the stratum rectangles to plot
    :param stratum_gap:     gap between strata in a group
    :param plot_height:     height of the generated plot
    :param plot_width:      width of the generated plot
    :param show_labels:     if True plots Stratum name with each Stratum

    :return:                matplotlib.Axes if ax is given else matplotlib.Figure, matplotlib.Axes
    """
    return_fig = False
    if not ax:

        fig, ax = plt.subplots()
        return_fig = True

    if not isinstance(data, pd.DataFrame):
        data = to_dataframe(x, alluvium, stratum, hue = hue)
        x = 'x'
        stratum = 'stratum',
        alluvium = 'alluvium'
        hue = 'hue'

    data = data.copy()
    if hue:
        for col in [stratum, hue]:
            data.loc[:, col] = data.loc[:, col].astype(str)

        data['grouping'] = data.loc[:, stratum]
        data.loc[:, stratum] = data[[stratum, hue]].agg('_'.join, axis = 1)

    else:
        data['grouping'] = data.loc[:, stratum]

    colors = get_color_dict(
        data,
        x,
        stratum,
        hue = hue,
        sns_palette = palette
    )

    strata, lodes, group_labels, groupings = aggregate_data(
        data,
        x,
        alluvium,
        stratum
    )

    plot_strata(
        strata,
        group_labels,
        groupings,
        colors,
        stratum_width,
        stratum_gap,
        ax,
        plot_height,
        plot_width,
        show_labels = show_labels
    )

    plot_flows(
        strata,
        lodes,
        ax
    )

    if not show_labels:
        ax.set_xticks([])

    ax.set_yticks([])
    for pos in ['top', 'bottom', 'left', 'right']:
        ax.spines[pos].set_visible(False)

    return ax if not return_fig else (fig, ax)